/* 
 *  $Id: mp7.c,v 1.1 2003/11/26 15:53:26 kogan Exp $
 *
 *  Copyright (c) 2003, Sergey Kogan
 *
 *  This is free software; you can redistribute it and/or modify it under the
 *  terms of the GNU General Public License as published by the Free Software
 *  Foundation; either version 2, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <PalmOS.h>
#include "../config.h"
#include "../dia.h"
#include "mp7.h"


#define MP7_CTL_CONNECT		DIA_CTL_USER+0
#define MP7_CTL_POLL		DIA_CTL_USER+1
#define MP7_CTL_DISCONNECT	DIA_CTL_USER+2
#define MP7_CTL_PING		DIA_CTL_USER+3
#define MP7_CTL_RECONNECT	DIA_CTL_USER+4
#define MP7_CTL_SETTIMING	DIA_CTL_USER+5

/*
 *  internal vars
 */

Boolean		_mp7Running		= false;
Boolean		_mp7Connected		= false;
Boolean		_mp7Polling		= false;
Boolean		_mp7ErrorMode		= false;
Boolean		_mp7Error		= false;	// TRUE if last operation was unsuccessful
UInt8		_mp7ErrorCounter	= 0;
Char		_mp7LastError[64];

Mp7WaitingFor	_mp7WaitingFor		= DIA_WAIT_NONE;
Mp7RequestType	_mp7Req;

DiaSampleType*	_mp7SampleP		= NULL;
Boolean		_mp7SampleReady		= false;

UInt16		_mp7PortId;
UInt16		_mp7AutoOff;
Boolean		_mp7PortOpen		= false;

UInt32		_mp7TicksPerSecond;
UInt32		_mp7LastReplyTicks	= 0;

/*
 *  API implementation
 */

#define MSEC2TICKS(a)	(((a)*_mp7TicksPerSecond)/1000)

/* internal functions decls */
DiaReturnType	_Mp7TryToSend() MP7_SECT;
DiaReturnType	_Mp7TryToRecv() MP7_SECT;
DiaReturnType	_Mp7OnIdle() MP7_SECT;
DiaReturnType	_Mp7ClosePort() MP7_SECT;
DiaReturnType	_Mp7OnRecv() MP7_SECT;
Int16		_Mp7Compare(const UInt8 *s1, const UInt8 *s2, UInt8 length) MP7_SECT;
DiaReturnType	_Mp7OpenPort(UInt32 baudRate, Boolean enableFlowControl) MP7_SECT;


inline void	_Mp7ClearErrors()
{
	_mp7Error = false;
}

/* start/stop diagnostic session */
DiaReturnType Mp7Start(UInt32 mask)
{
	if(_mp7Running) return DIA_OK;

	diaMod->mask = mask;

	_Mp7ClearErrors();
	_mp7ErrorCounter = 0;
	_mp7Running = true;
	_mp7Connected = false;
	_mp7Polling = false;
	_mp7WaitingFor = DIA_WAIT_NONE;
	_mp7PortOpen = false;
	_mp7TicksPerSecond = SysTicksPerSecond();
	_mp7SampleReady = false;

	_mp7AutoOff = SysSetAutoOffTime(0);

	return DIA_OK;
}

DiaReturnType Mp7Stop()
{
	if(!_mp7Running) return DIA_OK;

	if(_mp7Polling)
		Mp7StopSensorsPoll();	// stop polling

	// stop current request
	if(_mp7WaitingFor == DIA_WAIT_NONE || _mp7WaitingFor == DIA_WAIT_SEND)
		_mp7WaitingFor = DIA_WAIT_NONE;
	else
		while(1)
			if(Mp7Run() != DIA_OK) break;
	_mp7SampleReady = false;

	Mp7Ctl(MP7_CTL_DISCONNECT, NULL);	// close connection synchronously

//	if(_mp72PortOpen)
//		_Mp72ClosePort();

	if(_mp7AutoOff) {
		SysSetAutoOffTime(_mp7AutoOff);
		_mp7AutoOff = 0;
	}

	_mp7Running = false;

	return DIA_OK;
}

/* call this function as frequently as possible */
DiaReturnType Mp7Run()
{
	DiaReturnType return_event = DIA_OK;

	if(!_mp7Running) return DIA_OK;

	_mp7SampleReady = false;

	if(_mp7WaitingFor != DIA_WAIT_NONE)
	{
		if(_mp7WaitingFor == DIA_WAIT_SEND)
			return_event = _Mp7TryToSend();
		else
			return_event = _Mp7TryToRecv();
	}
	else
	// idle, check what we can do
	if(!_mp7Connected)
		return_event = Mp7Ctl(MP7_CTL_CONNECT, NULL);	// try to connect
	else
	if(_mp7Polling && diaMod->mask != 0)
		return_event = Mp7Ctl(MP7_CTL_POLL, NULL);
	else
		return_event = _Mp7OnIdle(); // send ping to ECU
	return return_event;
}

/* driver states */
Boolean Mp7IsRunning()
{
	return _mp7Running;
}

Boolean Mp7IsConnected()
{
	return _mp7Connected;
}

Boolean Mp7IsPolling()
{
	return _mp7Polling;
}

Char* Mp7GetLastError() /* return NULL if no errors */
{
	if(!_mp7Error) return NULL;
	return _mp7LastError;
}


/* polling */
DiaReturnType Mp7StartSensorsPoll()
{
	if(!_mp7Running) return DIA_ERROR;
	_mp7Polling = true;
	_mp7SampleReady = false;
	return DIA_OK;
}

DiaReturnType Mp7StopSensorsPoll()
{
	if(!_mp7Running) return DIA_ERROR;
	_mp7Polling = false;
	return DIA_OK;
}

DiaSampleType* Mp7GetSample()	/* return NULL if sample is not ready,
								flush sample */
{
#ifndef EMULATE_ECU_ANSWERS
	if(!_mp7Polling) return NULL;
	if(!_mp7SampleReady) return NULL;
	_mp7SampleReady = false;
#else
	_mp7SampleP = (DiaSampleType *)&_mp7Req.sample;
	_mp7SampleP->time = TimGetSeconds();
	_mp7SampleP->sensorPresent = 0xFFFFFFFF;
	_mp7SampleP->data[9]++;
#endif
	return _mp7SampleP;
}


/* operations */
DiaReturnType Mp7Ctl(UInt8 ctl, void *param)
{
	DiaCtlType ctlP = NULL;
	DiaReturnType r;

	if(_mp7WaitingFor != DIA_WAIT_NONE) return _Mp7Error("Mp7Ctl(): driver busy", errNone, false);

	_Mp7ClearErrors();
	_mp7SampleReady = false;

	ctlP = diaMod->ctl[ctl];
	if(ctlP == NULL)
		return _Mp7Error("null ctl called", errNone, false);

	_mp7Req.ctl = ctl;
	_mp7Req.ctl_param = param;

	r = ctlP(ctl, DIA_SEND, param);
	if(r != DIA_OK) return r;
	r =_Mp7TryToSend();
	if(r != DIA_OK) return r;
	r = _Mp7TryToRecv();
	if(r != DIA_OK) return r;

	return DIA_OK;
}

DiaReturnType _Mp7OnRecv()
{
	DiaCtlType ctlP = NULL;
	
	ctlP = diaMod->ctl[_mp7Req.ctl];
	return ctlP(_mp7Req.ctl, DIA_RECV, _mp7Req.ctl_param);
}

Boolean Mp7CtlGetResult(UInt8 *ctl, void **param,
						UInt16 *size)	/* return FALSE if result
										is not ready */
{
	return _Mp7Error("Not implemented", errNone, true);
}


/*** internal functions implementation ***/

DiaReturnType _Mp7TryToSend()
{
    if(_mp7WaitingFor != DIA_WAIT_SEND) return DIA_OK;

	if((TimGetTicks() - _mp7LastReplyTicks) >= (MSEC2TICKS(diaMod->timing->minP3)))
	{
		{
			Err err;

			SrmReceiveFlush(_mp7PortId, 0);
			SrmSend(_mp7PortId, _mp7Req.send, _mp7Req.send_size, &err);
			if(err != errNone)
				return _Mp7Error("SrmSend()", err, true);

			_mp7Req.send_ticks = TimGetTicks();
		}

		_mp7WaitingFor = DIA_WAIT_ECHO;
	}
	return DIA_OK;
}

Int16 _Mp7Compare(const UInt8 *s1, const UInt8 *s2, UInt8 length)
{
	UInt8 i;
	Int16 r = 0;
	for(i = 0; i < length; i++)
		if((r = *s1++ - *s2++)) return r;
	return r;
}


DiaReturnType _Mp7TryToRecv()
{
	Err err;
	UInt32 bytes;
	UInt32 bytes_expected;
	UInt32 timeout;

	switch(_mp7WaitingFor)
	{
	case DIA_WAIT_ECHO:
		bytes_expected = _mp7Req.send_size;
		timeout = MSEC2TICKS(diaMod->timing->maxP2);
		break;
	case DIA_WAIT_HDR:
		bytes_expected = 4;
		timeout = MSEC2TICKS(diaMod->timing->maxP2 + 4*diaMod->timing->maxP1);
		break;
	case DIA_WAIT_REPLY:
		bytes_expected = _mp7Req.recv_size + 1;
		timeout = MSEC2TICKS(diaMod->timing->maxP2 + (bytes_expected + 4)*diaMod->timing->maxP1);
		break;
	default:
		return DIA_OK;
	}

	err = SrmReceiveCheck(_mp7PortId, &bytes);
	if(err != errNone)
		return _Mp7Error("SrmReceiveCheck()", err, true);

	if(bytes >= bytes_expected)
	{
		switch(_mp7WaitingFor)
		{
		case DIA_WAIT_ECHO:
			// recv echo
			SrmReceive(_mp7PortId, _mp7Req.recv, bytes_expected, 1, &err);
			if(err != errNone)
				return _Mp7Error("SrmReceive()", err, true);
			// compare
			if(_Mp7Compare(_mp7Req.send, _mp7Req.recv, _mp7Req.send_size) != 0)
				return _Mp7Error("Invalid echo", errNone, true);
			// go to the next step
			_mp7WaitingFor = DIA_WAIT_HDR;
			return _Mp7TryToRecv();
		case DIA_WAIT_HDR:
			// get packet length
			SrmReceive(_mp7PortId, _mp7Req.recv, 3, 1, &err);
			if(err != errNone)
				return _Mp7Error("SrmReceive()", err, true);

			_mp7Req.recv_size = 0x3F & (_mp7Req.recv_checksum = _mp7Req.recv[0]);
			_mp7Req.recv_checksum += _mp7Req.recv[1];
			_mp7Req.recv_checksum += _mp7Req.recv[2];

			if(_mp7Req.recv_size == 0)
			{
				SrmReceive(_mp7PortId, _mp7Req.recv, 1, 1, &err);
				if(err != errNone)
					return _Mp7Error("SrmReceive()", err, true);
				_mp7Req.recv_checksum += _mp7Req.recv[0];
				_mp7Req.recv_size = _mp7Req.recv[0];
			}
			// go to the next step
			_mp7WaitingFor = DIA_WAIT_REPLY;
			return _Mp7TryToRecv();
		case DIA_WAIT_REPLY:
			// read reply
			SrmReceive(_mp7PortId, _mp7Req.recv + _mp7Req.recv_to, _mp7Req.recv_size+1, 1, &err);
			if(err != errNone)
				return _Mp7Error("SrmReceive()", err, true);

			// store reply ticks
			_mp7LastReplyTicks = TimGetTicks();

			// check checksum
			{
				UInt8 i;
				for(i = 0; i < _mp7Req.recv_size; i++)
					_mp7Req.recv_checksum += _mp7Req.recv[_mp7Req.recv_to+i];
			}
			if(_mp7Req.recv_checksum != _mp7Req.recv[_mp7Req.recv_to+_mp7Req.recv_size])
				return _Mp7Error("Bad checksum", errNone, true);

			// check answer
			if(_mp7Req.recv[_mp7Req.recv_to+0] != (_mp7Req.send[3] | 0x40))
			{

				if(_mp7Req.recv[_mp7Req.recv_to+0] != 0x7F)
					return _Mp7Error("Unknown answer", errNone, true);
				if(_mp7Req.recv_size != 3)
					return _Mp7Error("Bad err answer", errNone, true);
				if(_mp7Req.recv[_mp7Req.recv_to+1] != _mp7Req.send[3])
					return _Mp7Error("Bad err answer ID", errNone, true);

				switch(_mp7Req.recv[_mp7Req.recv_to+2])
				{
				case 0x10:
					return _Mp7Error("ECU: General reject", errNone, false);
				case 0x11:
					return _Mp7Error("ECU: Service not supported", errNone, false);
				case 0x12:
					return _Mp7Error("ECU: Subfunction not supported", errNone, false);
				case 0x31:
					return _Mp7Error("ECU: Request out of range", errNone, false);
				case 0x72:
					return _Mp7Error("ECU: Transfer aborted", errNone, false);
				case 0x77:
					return _Mp7Error("ECU: Block checksum error", errNone, false);
				case 0x21: // busy - retry later
					_mp7Req.retry++;
					_mp7WaitingFor = DIA_WAIT_SEND;
					return _Mp7TryToSend();
				}
		   	}

			// clear WaitingFor
			_mp7WaitingFor = DIA_WAIT_NONE;

			_mp7ErrorCounter = 0;

			// call OnRecv();
			return _Mp7OnRecv();
		default:
			return DIA_ERROR;
		}
	}
	else
		if(TimGetTicks() - _mp7Req.send_ticks > timeout)
		{
			if(_mp7Req.retry >= 15) 
				return _Mp7Error("Recv timeout", errNone, false);
			else
			{
				_mp7Req.retry++;
//				if(bytes > 0)
//					SrmReceiveFlush(_mp7PortId, 0);
				_mp7WaitingFor = DIA_WAIT_SEND;
				return _Mp7TryToSend();
			}
		}

	return DIA_OK;
}

DiaReturnType _Mp7OnIdle()
{
	if(_mp7WaitingFor == DIA_WAIT_NONE &&
		((TimGetTicks() - _mp7LastReplyTicks) > MSEC2TICKS(diaMod->timing->maxP3 >> 1)))
		return Mp7Ctl(MP7_CTL_PING, NULL);
	return DIA_OK;
}

DiaReturnType _Mp7Error(Char *msg, Err err, Boolean fatal) // always return true
{
	// check error mode
	if(_mp7ErrorMode) return DIA_ERROR;

	// enter error mode
	_mp7ErrorMode = true;

	// reset internal states
	_mp7WaitingFor = DIA_WAIT_NONE;

//	if(fatal)
//		Mp7Ctl(MP7_CTL_DISCONNECT, NULL);
	_mp7ErrorCounter++;
	if(_mp7ErrorCounter > 6)
	{
 		_mp7ErrorCounter = 0;
		Mp7Ctl(MP7_CTL_RECONNECT, NULL);
	}

	// set error
	StrCopy(_mp7LastError, msg);
	if(err != errNone)
	{
		Char buf[64];

		if(SysErrString(err, buf, 64) != NULL)
		{
			StrCat(_mp7LastError, ": ");
			StrCat(_mp7LastError, buf);
		}
	}
	_mp7Error = true;

	if(_mp7PortOpen)
	{
		// flush input/output buffer
		SrmReceiveFlush(_mp7PortId, 0);
		SrmSendFlush(_mp7PortId);
		// clear Srm errors
		SrmClearErr(_mp7PortId);
	}

	// leave error mode
	_mp7ErrorMode = false;
	return DIA_ERROR;//_Mp7Error("Not implemented", errNone, true);
}

DiaReturnType _Mp7SendAsync(UInt8 *req, UInt8 size, UInt8 recv_to)
{
	UInt8 checksum = 0;
	UInt8 i;

	if(size > 63)
		return _Mp7Error("_Mp7SendAsync(): Can't send long packet", errNone, false);

	checksum += _mp7Req.send[0] = 0x80 + size;
	checksum += _mp7Req.send[1] = 0x11;
	checksum += _mp7Req.send[2] = 0xF1;
	for(i = 0; i < size; i++)
		checksum += _mp7Req.send[i+3] = req[i];
	_mp7Req.send[size+3] = checksum;
	_mp7Req.send_size = size+4;
	_mp7Req.recv_to = recv_to;

	_mp7Req.retry = 0;
	_mp7WaitingFor = DIA_WAIT_SEND;

	return DIA_OK;
}

DiaReturnType _Mp7OpenPort(UInt32 baudRate, Boolean enableFlowControl)
{
	Err err;

	if(_mp7PortOpen) return DIA_OK;

	err = SrmOpen(serPortCradleRS232Port, baudRate, &_mp7PortId);
	if(err != errNone)
		err = SrmOpen(serPortCradlePort, baudRate, &_mp7PortId);
	if(err != errNone)
		return _Mp7Error("SrmOpen()", err, true);

	_mp7PortOpen = true;
		
	if(!enableFlowControl) {
		Int16 paramSize;
		UInt32 flags =	srmSettingsFlagBitsPerChar8 | srmSettingsFlagStopBits1;
		paramSize = sizeof(flags);
		err = SrmControl(_mp7PortId, srmCtlSetFlags, &flags, &paramSize);
		if(err != errNone)
			return _Mp7Error("SrmControl()", err, true);
	}

	return DIA_OK;
}

DiaReturnType _Mp7ClosePort()
{
	Err err;

	if(!_mp7PortOpen) return DIA_OK;

	_mp7PortOpen = false;
	err = SrmClose(_mp7PortId);
	if(err != errNone)
		return _Mp7Error("SrmClose()", err, false);

	return DIA_OK;
}

/*
 * ctl handlers
 */

DIA_DEFINE_CTL(Mp7Connect)
{
	if(op == DIA_SEND)
	{
		DiaReturnType r;
		Err err;
		UInt8 zero = 0;
//  	UInt16 pulse_ticks;
		UInt32 rate = 10400;
		UInt16 rate_size = sizeof(rate);
		UInt8 req[] = { 0x81 };

		if(_mp7Connected) return DIA_RESULT;

		r = _Mp7OpenPort(360, false);
		if(r != DIA_OK) return r;

		// do fast init
		// send 25ms pulse
		err = SrmSend(_mp7PortId, &zero, 1, &err);
		if(err != errNone)
			SrmClearErr(_mp7PortId);

		err = SrmReceive(_mp7PortId, &zero, 1, 10, &err);
    	if(err != errNone)
    		SrmClearErr(_mp7PortId);

		err = SrmControl(_mp7PortId, srmCtlSetBaudRate, &rate, &rate_size);
		if(err != errNone)
			return _Mp7Error("SrmControl()", err, true);

		return _Mp7SendAsync(req, 1, 0);
	}
	else
	{
		// check result
		if(_mp7Req.recv_size != 3)
			return _Mp7Error("Inv. answer length", errNone, true);
		if(_mp7Req.recv[1] != 0xEF || _mp7Req.recv[2] != 0x8F)
			return _Mp7Error("Inv. answer bytes", errNone, true);

		_mp7Connected = true;

		return Mp7Ctl(MP7_CTL_SETTIMING, NULL);
	}
}

DIA_DEFINE_CTL(Mp7SetTiming)
{
	if(op == DIA_SEND)
	{
		UInt8 req[] = { 0x83, 0x03, 0x00, 0x50, 0x01, 0x14, 0x00 };
		return _Mp7SendAsync(req, 7, 0);
	}
	else
	{
		// check result
		if(_mp7Req.recv_size != 2)
			return _Mp7Error("Inv. answer length", errNone, true);
//		if(_mp72Req.recv[1] != 0xEF || _mp72Req.recv[2] != 0x8F)
//			return _Mp72Error("Inv. answer bytes", errNone, true);
		return DIA_RESULT;
	}
}

DIA_DEFINE_CTL(Mp7Disconnect)
{
	if(op == DIA_SEND)
	{
		UInt8 req[] = { 0x82 };

		// do disconnect
		if(_mp7Connected)
		{
			_Mp7SendAsync(req, 1, 0);

			while(1)
				if(Mp7Run() != DIA_OK) break;

			_mp7Connected = false;
		}

		_Mp7ClosePort();
	}
	return DIA_RESULT;
}

DIA_DEFINE_CTL(Mp7Reconnect)
{
	if(op == DIA_SEND)
	{
		// do disconnect
		_mp7Connected = false;
		_Mp7ClosePort();
	}
	return DIA_RESULT;
}

DIA_DEFINE_CTL(Mp7Ping)
{
	if(op == DIA_SEND)
	{
		UInt8 req[] = { 0x3e };
		return _Mp7SendAsync(req, 1, 0);
	}
	return DIA_OK;
}

DIA_DEFINE_CTL(Mp7ReadIdent)
{
	return _Mp7Error("Not implemented", errNone, true);
}

DIA_DEFINE_CTL(Mp7ClearErrors)
{
	return _Mp7Error("Not implemented", errNone, true);
}

DIA_DEFINE_CTL(Mp7ReadErrors)
{
	return _Mp7Error("Not implemented", errNone, true);
}

DIA_DEFINE_CTL(Mp7EcuReset)
{
	return _Mp7Error("Not implemented", errNone, true);
}

